from ply import lex

tokens = [
    "TABLE",
    "ARRTABLE",
    "ID",
    "STR",
    "MLSTR",
    "DATE",
    "TIME",
    "DATETIME",
    "INT",
    "HEX",
    "BIN",
    "OCT",
    "FLOAT",  # need to implement exponents check https://toml.io/en/
    "BOOL",
    "INF",
    "NAN",
    "COMMENT",
    "NEWLINE",
]

states = (('array','inclusive'),
          ('dict','inclusive'))

def t_TABLE(t):
    r"\[\s*(([\w_-]+|\"[\w_-]+\"|\'[\w_-]+\')(\s*\.\s*([\w_-]+|\"[\w_-]+\"|\'[\w_-]+\'))*)\s*\]"
    header = t.lexer.lexmatch.groups()[1]
    header = [s.strip(" \"\'") for s in header.split('.')]
    t.value = header
    return t 

def t_ARRTABLE(t):
    r"\[\s*\[\s*(([\w_-]+|\"[\w_-]+\"|\'[\w_-]+\')(\s*\.\s*([\w_-]+|\"[\w_-]+\"|\'[\w_-]+\'))*)\s*\]\s*\]"
    header = t.lexer.lexmatch.groups()[6]
    header = [s.strip(" \"\'") for s in header.split('.')]
    t.value = header
    return t 

def t_INITIAL_NEWLINE(t):
    r"\n+\s*"
    return t

def t_INITIAL_open_array(t):
    r"\["
    t.type = "["
    t.lexer.push_state('array')
    return t

def t_array_close_array(t):
    r"\]"
    t.type = "]"
    t.lexer.pop_state()
    return t 

def t_INITIAL_open_dict(t):
    r"\{"
    t.type = "{"
    t.lexer.push_state('dict')
    return t

def t_dict_close_dict(t):
    r"\}"
    t.type = "}"
    t.lexer.pop_state()
    return t 

# needs to check if datetime is valid
def t_DATETIME(t):
    r"\d{4}-\d{2}-\d{2}[T ]\d{2}:\d{2}:\d{2}(\.\d{1,6})?(Z|[+-]\d{2}:\d{2})"
    return t


# needs to check if date is valid
def t_DATE(t):
    r"\d{4}-\d{2}-\d{2}"
    return t


# needs to check if time is valid
def t_TIME(t):
    r"\d{2}:\d{2}:\d{2}(\.\d{1,6})?"
    return t


# needs number grouping (example : flt8 = 224_617.445_991_228)
def t_FLOAT(t):
    r"[+-]?\d+(_\d+)*\s*\.\s*\d+(_\d+)*([eE][-+]?\d+(_\d+)*)?"
    #case where float appears on the left side with spaces in between
    if t.value.__contains__(' '):
        t.type = "ID"
        t.value = [s.strip(' ')  for s in t.value.split('.')]
    return t


# needs number grouping (example : int6 = 5_349_221)
def t_INT(t):
    r"[-+]?(\d+(_\d+)*)"
    return t


# needs number grouping (example : hex3 = 0xdead_beef)
def t_HEX(t):
    r"0x[0-9a-fA-F]+(_[0-9a-fA-F]+)*"
    return t


def t_BIN(t):
    r"0b[01]+(_[01]+)*"
    return t


def t_OCT(t):
    r"0o[0-7]+(_[0-7]+)*"
    return t


def t_INF(t):
    r"[-+]?inf"
    return t


def t_NAN(t):
    r"[+-]?nan"
    return t


# booleans are always lowercase
def t_BOOL(t):
    r"(true|false)"
    return t


# ID needs to be the last so it doesnt catch everything (literally)
def t_ID(t):
    r"(([\w_]+)|(\"[\w_]+\"|\'[\w_]+\')\s*\.\s*([\w_]+|\"[\w_]+\"|\'[\w_]+\'))(\s*\.\s*([\w_]+|\"[\w_]+\"|\'[\w_]+\'))*"
    t.value = [s.strip(" \"'") for s in t.value.split('.')]
    return t


def t_MLSTR(t):
    r"(\"\"\"(?:[^\"\\]|\\.)*\"\"\")|(\'\'\'.*\'\'\')"
    t.value = t.value.strip("\"'")
    return t


# STR needs to be the first one to catch
def t_STR(t):
    r"(\"(?:[^\"\\]|\\.)*\")|(\'[^\']*\')"
    t.value = t.value.strip("\"'")
    return t


def t_COMMENT(t):
    r"\#.*"
    pass


def t_error(t):
    print("Illegal character '%s'" % t.value[0])
    t.lexer.skip(1)


t_ignore = "\t "
t_array_dict_ignore = "\t\n "

literals = ",="

lexer = lex.lex()
